Require Import Omega. 
Require Import Arith.
Require Import NPeano.
Require Import List.
Require Import LIFO.List.InSig.
Require Import LIFO.List.Misc.
Require Import LIFO.List.SeqNat.
Require Import LIFO.List.Length.
Require Import LIFO.List.Flatten.
Require Import LIFO.List.Notations.
Require Import LIFO.Nat.BoundedNat.
Require Import LIFO.Nat.ProofEquality.
Require Import CertifiedBSML.Primitives.Specification.

Set Implicit Arguments.

Module Type TYPE  
  (BsmlPrimitives: PRIMITIVES).

  Import BsmlPrimitives.

  Definition lmkpar (A:Type)(f:processor->A) :
    forall (i : processor),   get (proj1_sig (mkpar f)) i = f i 
      := 
      proj2_sig (mkpar f).

  Definition lapply(A B:Type)(vf:par (A->B))(vv:par A):
    forall (i : processor), 
      get (proj1_sig (apply vf vv)) i  =
      get vf i (get vv i)
    :=proj2_sig (apply vf vv).

  Definition lproj(A:Type)(v:par A):
    forall (i : processor), proj1_sig (proj v) i = get v i 
    :=  proj2_sig (proj v).

  Definition lput(A:Type)(v:par(processor->A)) : 
    forall (i j : processor),
                        get (proj1_sig (put v)) i j = get v j i 
     := proj2_sig (put v).

  Hint Rewrite lmkpar lapply lproj lput : bsml.

  Lemma bsp_pSMaxPid: 
    exists maxPid:nat, bsp_p = S maxPid.
  Proof.
    case_eq(bsp_p). 
       intro Hbsp_p; set(Hp:= bsp_pLtZero); rewrite Hbsp_p in *; contradict Hp; omega.
       intro maxPid; exists maxPid; trivial.
  Qed.

  Lemma bsp_pNotZero: bsp_p<>0.
  Proof.
    destruct(bsp_pSMaxPid); intuition.
  Qed.

  Hint Resolve bsp_pLtZero bsp_pNotZero: bsml.


  (** Higher processor number  *)
  Definition max_pid:nat.
  case_eq(bsp_p). 
  intro Hbsp_p; set(Hp:= bsp_pNotZero); rewrite Hbsp_p in *; contradict Hp; omega.
  intros maxPid H. exact maxPid.
  Defined.

  Lemma max_pid_S_bsp_p: bsp_p = S max_pid.
  Proof. 
    unfold    max_pid.
    pattern bsp_pNotZero.
    match goal with |- ?H  ?lem => assert (H' : forall lem, H lem) end.
    pattern bsp_p.
    match goal with |- ?H  ?lem => assert (H' : forall lem, H lem) end.
    intros lem lem0.    
    destruct lem;[elim lem0|]; reflexivity.
    generalize (H' bsp_p).
    intros H lem.   
    apply (H lem).
    apply (H' bsp_pNotZero).
  Qed.

  Program Definition firstProcessor : processor := 0.
  Next Obligation.
    destruct(bsp_pSMaxPid) as [maxPid HmaxPid];
      rewrite HmaxPid; 
        intuition.
  Qed.

  Program Definition lastProcessor : processor := bsp_p - 1.
  Next Obligation.
    destruct(bsp_pSMaxPid) as [maxPid HmaxPid]; rewrite HmaxPid; omega.
  Qed.

  Program Definition 
    processorOfNatMod (n:nat): processor := 
    n mod bsp_p.
  Next Obligation.
    apply mod_upper_bound;
      destruct(bsp_pSMaxPid) as [maxPid HmaxPid]; rewrite HmaxPid; omega.
  Defined.

  Program Definition   natOfProcessor (p:processor) :nat :=  p.

  Definition processorOfNat i H := (exist (fun i => i < bsp_p)  i H).
  Implicit Arguments processorOfNat [ ].
  
  Hint Unfold listProj: bsml_properties.
  Hint Rewrite inSigProj : bsml_properties.

  Lemma processorOfNatModLtBspPId : 
    forall(n:nat)(H:n < bsp_p),
      processorOfNatMod n = exist (fun n=>n<bsp_p) n H.
  Proof.
    intros n H;
      unfold processorOfNatMod;
        apply proj1_sigBoundedNatInjective; simpl;
          apply NZDivP.mod_small; omega.
  Qed.

  Hint Resolve processorOfNatModLtBspPId :bsml.


  (** equality on processors is decidable  *)
Lemma processorEqDec : forall (p1 p2:processor), {p1=p2}+{p1<>p2}.
  Proof.
    intros p1 p2. 
    destruct p1 as [p1 p1_proc] ; destruct p2 as [p2 p2_proc].
    
    destruct (eq_nat_dec p1 p2) as [peq | diff].
        revert p1_proc; rewrite peq ; intros p1_proc.
        rewrite (ProofEquality.ltUniquenessProof p2 bsp_p  p1_proc p2_proc ).
        left; reflexivity.  

        right.
        intro abs. inversion abs.
        contradiction.
Qed.

  (** equality on processors is is equality on index  *)
  (* Processor are equal if their natural pid are equal *)
    Lemma processor_eq : forall (i j: processor),
      proj1_sig i = proj1_sig j -> i = j.
    Proof.
      intros. destruct i, j. 
      simpl in H. subst.
      replace l with l0 by apply ltUniquenessProof.
      reflexivity.
    Qed.

    Program Definition processor_beq (p1 p2 : processor) := EqNat.beq_nat p1 p2.

  Lemma processorLtBspP:
    forall i:processor, proj1_sig i < bsp_p.
  Proof.
    destruct i; auto.
  Qed.

  Hint Resolve processorLtBspP :bsml.


Lemma lastProcessorBspp : S (proj1_sig lastProcessor) = bsp_p.
Proof.
  simpl; generalize bsp_pNotZero; intros. omega.
Qed.
Hint Resolve lastProcessorBspp : bsml. 

Lemma bsppLastProcessor :bsp_p = S (proj1_sig lastProcessor). auto with bsml. Qed.

  Program Definition processors : { l : list processor | listProj l = seq 0 bsp_p} := 
    inSig (fun i=>i<bsp_p) (seq 0 bsp_p) (inSeq0Lt bsp_p). 
  Next Obligation.
    autorewrite with bsml_properties; auto.
  Qed.

  Lemma processorsNatSeq : 
      map natOfProcessor (proj1_sig processors) = seq 0 bsp_p.
    Proof.
      rewrite <- (proj2_sig processors).
      unfold listProj, natOfProcessor.
      reflexivity.      
    Qed.

  Program Definition processorsLength: 
    List.length processors = bsp_p.
  Proof.
    unfold processors; autorewrite with length; unfold processor; 
      autorewrite with length; trivial.
  Qed.

  Hint Rewrite processorsLength : length.

  Program Definition inProcessors : 
    forall i:processor, In i processors.
  Proof.
    intro i; unfold processors; simpl; destruct i; 
      apply inInSig1; apply inSeq; split; omega.
  Qed.

  Program Definition nthProcessors : 
    forall (n:processor) d, nth n  processors d = n.
  Proof.
    intros n d; unfold processors; destruct n; simpl. 
    destruct d. rewrite (nthInSigPI _ x _  l).
    reflexivity.    
    rewrite seq_nth.
    reflexivity.    
    omega.
  Qed.
  
  Hint Resolve  inProcessors :bsml.

  Lemma processorsBeginsWithFirstProcessor:
      proj1_sig processors = firstProcessor :: (tail (proj1_sig processors)).
  Proof.
    destruct bsp_pSMaxPid as [maxPid HmaxPid].
    assert(seq 0 bsp_p = 0::seq 1 maxPid) as Hseq by (rewrite HmaxPid; trivial).
    pose(Hlen := processorsLength); generalize  Hlen; clear Hlen; intro Hlen.
    case_eq (proj1_sig processors). 
      intro Hdef; contradict Hlen; rewrite Hdef, HmaxPid; simpl; omega.
      intros first tail Hdef; 
        destruct processors as [processorList Hprocessors]; simpl in Hdef;
          clear Hlen; rewrite Hdef in Hprocessors; simpl in Hprocessors; 
            rewrite Hseq in Hprocessors; inversion Hprocessors;
              f_equal; apply proj1_sigBoundedNatInjective;
                assumption.
  Qed.

  Lemma splitProcessors (p:processor) : 
    exists l1, exists l2, (proj1_sig processors) = l1 ++ (p::l2).
  Proof.
     apply In_split.
    apply inProcessors.
  Qed.

Global  Ltac destructProcessors p :=
    let lg:= fresh "lg" in let ld:= fresh "ld" in let H:= fresh "processors_decomp" in 
      destruct (splitProcessors p) as [lg H]; destruct H as [ld H]; rewrite H . 

  Lemma splitedProcessorsLength  l1 p l2 :
    (proj1_sig processors) = l1 ++ p :: l2 -> length l1 + length l2 + 1 = bsp_p.
  Proof.
    intros.
    apply f_equal with (f:=length(A:=processor)) in H.
    rewrite processorsLength in H. rewrite H.
    rewrite app_length. simpl. omega.
  Qed.

  Lemma splitedProcessorsLengthLeft l1 p l2 :
    proj1_sig processors = l1 ++ p :: l2 -> 
    proj1_sig p = length l1.
  Proof.
    intros.
    pose (Hlen := splitedProcessorsLength l1 p l2 H). 
    generalize Hlen; clear Hlen; intro.
    apply f_equal with (f:=map natOfProcessor ) in H.
    rewrite processorsNatSeq in H.
    apply f_equal with (f:=fun x => nth (length l1) x 0) in H.
    rewrite seq_nth in H.
    simpl in H.
    change 0 with (natOfProcessor firstProcessor) in H.
    rewrite map_nth in H.
    rewrite app_nth2 in H. rewrite <- minus_n_n in H.
    simpl in H. symmetry. assumption.
    auto. omega.
  Qed.

  Require Import LIFO.List.Firstn_skipn.

  Lemma splitedProcessorsLeftInf  l1 p l2 :
    (proj1_sig processors) = l1 ++ p :: l2 -> 
    forall p', 
      In p' l1 ->natOfProcessor p' < natOfProcessor p.
  Proof.
    intros.
    assert (p_length_l1 := splitedProcessorsLengthLeft  l1 p l2 H).
    (* pose (Hlen := split_all_processors_lengths l1 p l2 H). generalize Hlen; clear Hlen; intro. *)
    apply f_equal with (f:=map natOfProcessor) in H.
    rewrite processorsNatSeq in H.
    rewrite map_app in H.  
   apply f_equal with (f:=fun x => firstn (length l1) x) in H.
    replace (bsp_p) with ((natOfProcessor p) +(bsp_p -(natOfProcessor p))) in H.
    Focus 2.
         assert (natOfProcessor p < bsp_p).
         destruct p;simpl; assumption.
         omega.
    rewrite seqApp0 in H.     
     rewrite  (firstn_app_length (seq 0 (natOfProcessor p))) in H;[|
         autorewrite with length; auto].
    rewrite (firstn_app_length (map natOfProcessor l1)) in H;[|
         autorewrite with length; auto]. 
    apply inSeq0Lt.
    rewrite H.    
    apply in_map.
    assumption.    
  Qed.

  Lemma splitedProcessorsRightSup l1 p l2 :
    (proj1_sig processors) = l1 ++ p :: l2 -> forall p', In p' l2 ->natOfProcessor p < natOfProcessor p'.
  Proof.
    intros.
    assert (p_length_l1 := splitedProcessorsLengthLeft  l1 p l2 H).    
    (* pose (Hlen := split_all_processors_lengths l1 p l2 H). generalize Hlen; clear Hlen; intro. *)
    apply f_equal with (f:=map natOfProcessor) in H.
    rewrite processorsNatSeq in H.
    rewrite map_app in H.  
   apply f_equal with (f:=fun x => skipn (length l1) x) in H.
    replace (bsp_p) with ((natOfProcessor p) +(bsp_p -(natOfProcessor p))) in H.
    Focus 2.
         assert (natOfProcessor p < bsp_p).
         destruct p;simpl; assumption.
         omega.
    rewrite seqApp0 in H.     
    rewrite  (skipn_app_length (seq 0 (natOfProcessor p))) in H;[|
         autorewrite with length; auto].
    rewrite (skipn_app_length (map natOfProcessor l1)) in H;[|
         autorewrite with length; auto].
    change (natOfProcessor p < natOfProcessor p') with ( S (natOfProcessor p) <= natOfProcessor p').
    destruct (eq_nat_dec 0 (bsp_p - natOfProcessor p) ).
    rewrite <- e in H.    
    simpl in H.
    discriminate.
    assert ( H1 : bsp_p - natOfProcessor p = S (bsp_p - natOfProcessor p -1))by omega. 
    rewrite H1 in H.    
    simpl seq in H.
    injection H.
    intros H2. 
    apply (inSeqGt (S (natOfProcessor p))  (bsp_p - natOfProcessor p - 1)).
    rewrite H2.
    apply in_map.
    assumption.    
  Qed.


  Program Definition processorsUpTo (i:processor) : list processor := 
    inSig (fun n:nat=>n<bsp_p) (seq 0 (S i)) _.
  Next Obligation.
    destruct H; [
      subst; auto with bsml |
      destruct i as [i Hi]; simpl in *; apply inSeqLt in H; omega ].
  Defined.

  Program Definition processorsUpToLastProcessorIsProcessors: 
    processorsUpTo lastProcessor = processors.
  Proof.
    unfold processorsUpTo, lastProcessor, processors.
    replace (proj1_sig(exist (fun l : list processor => listProj l = seq 0 bsp_p)
       (inSig (fun i : nat => i < bsp_p) (seq 0 bsp_p) (inSeq0Lt bsp_p))
       processors_obligation_1)) with 
    (inSig (fun i : nat => i < bsp_p) (seq 0 bsp_p) (inSeq0Lt bsp_p)) by auto.
    apply inSig_eq with (l2:=seq 0 bsp_p); simpl.
    destruct bsp_pSMaxPid as [m Hm]; rewrite Hm; simpl; rewrite <- minus_n_O; trivial.
  Qed.
  

  Lemma processorsUpToApp:
    forall (n:nat)(H:S n<bsp_p), 
      let processorSn := exist (fun n=>n<bsp_p) (S n) H in
        let processorN := exist (fun n=>n<bsp_p) n (Lt.lt_trans n (S n) bsp_p (lt_n_Sn n) H) in
          processorsUpTo processorSn = (processorsUpTo processorN) ++ (processorSn::nil).
  Proof.
    intros n H processorSn processorN; unfold processorsUpTo.
    assert(forall n0, In n0 (seq (S n) 1) -> n0 < bsp_p) as Hn by
      (intros n' H0; destruct H0; elim H0; assumption).
    assert(forall n0, In n0 ((seq 0 (S (proj1_sig processorN))) ++ (seq (S n) 1)) -> n0<bsp_p) as H3 by
      (intros n0 H0;simpl in H0; destruct H0; [
         rewrite <- H0; apply bsp_pLtZero |
         apply in_app_or in H0; destruct H0 as [H0 | H0]; [
           apply inSeqLt in H0; omega |
           simpl in H0; destruct H0; elim H0; assumption ] ]).
    assert(forall n':nat, In n' (seq 0 (S(S n))) -> n' < bsp_p) as H3' by
      (intros n' H0; apply H3; 
        rewrite <- seqAppN; rewrite plus_comm; assumption).
    replace (processorSn::nil) with (inSig _ (seq (S n) (S O)) Hn).
      rewrite inSig_app with (H3:=H3);
        rewrite inSig_eq with (H1:=H3)(H2:=H3')(l2:=seq 0 (S(S n))).
      apply inSig_eq; trivial.
      rewrite <- seqAppN; rewrite plus_comm; trivial.
      simpl; f_equal; apply proj1_sigBoundedNatInjective; trivial.
  Qed.

  Program Definition processorsStartingAt (i:processor) : list processor := 
    inSig (fun n:nat=>n<bsp_p) (seq i (bsp_p-i)) _.
  Next Obligation.
    apply inSeqLt in H.
    destruct i.    simpl in *. 
    omega.
  Defined.


  Program Definition processorsStartingAtFirstProcessorIsProcessors: 
    processorsStartingAt firstProcessor = processors.
  Proof.
    unfold processorsStartingAt, firstProcessor, processors.
    simpl.
    apply inSig_eq with (l2:=seq 0 bsp_p); simpl.
    rewrite <- minus_n_O; trivial.
  Qed.

  Program Definition processorsRevInduction : 
    forall P,
      (forall (p Sp:processor) ,p< (bsp_p-1)-> Sp= S p -> P (processorsStartingAt Sp) -> P (processorsStartingAt p))
      -> P (processorsStartingAt lastProcessor)
      -> P ( processors):=_.
Next Obligation.  
  destruct p; simpl in *;  omega.
Qed.

Obligation Tactic :=idtac.
Next Obligation.
pattern lastProcessor.
match goal with 
  [|- ?P lastProcessor] => let H:=fresh "H" in assert (H: forall p, P p);[|simpl;apply (H lastProcessor)]
end.
intros p P X X0.
rewrite <- processorsStartingAtFirstProcessorIsProcessors.
destruct p.
induction x.
assert ((exist (fun pid : nat => pid < bsp_p) 0 l) = firstProcessor).
apply processor_eq. reflexivity.
rewrite H in X0.
assumption.
assert ( x < bsp_p) by omega.
pose (p:=(exist (fun pid : nat => pid < bsp_p) (x) H)).
pose (Sp:=(exist (fun pid : nat => pid < bsp_p) (S x) l)).
assert (Hp: proj1_sig p < bsp_p -1) by (simpl in *;omega).
assert (H':  Sp =  exist (fun pid : nat => pid < bsp_p) (S (proj1_sig p)) (processorsRevInduction_obligation_1 P p Sp Hp) ).
apply processor_eq. reflexivity.
apply (IHx _  (X _ _ _ H'  X0)).
Qed.

Program Definition listOfParList (A:Type)(v:par (list A)) : list A := 
    flatten (map (get v) processors).

  Ltac listOfParList_processorsUpTo :=
    unfold listOfParList;
      rewrite <- processorsUpToLastProcessorIsProcessors;
        pattern lastProcessor;
          match goal with 
            [|- ?P lastProcessor] => let H:=fresh "H" in assert (H: forall p, P p);[|simpl;apply (H lastProcessor)]
          end.

  Ltac listOfParListInduction :=
    listOfParList_processorsUpTo;
    let p:=fresh "p" in let pProc:=fresh "pProc" in 
      intro p; 
        destruct p as [p pProc];
          induction p;      (* induction sur le dernier processeur *)
            [               (* pour avoir un joli get sur firstProcessor *)
              intros; 
              repeat    
                match goal with 
                  [|-context [Flatten.flatten(map (get ?plst) (processorsUpTo ?p))] ]=>
                  replace (Flatten.flatten(map (get plst) (processorsUpTo p))) with
                    (get plst firstProcessor)
                    by
                    (simpl;rewrite app_nil_r;   f_equal;  apply processor_eq; reflexivity);
                    try (replace (p) with (firstProcessor) in * by (apply processor_eq; reflexivity))
                end 
              |
              try (  (* séparation du dernier processeur de la liste de processeurs *)
                intros;
                rewrite processorsUpToApp;
                  try                 (* pour avoir de jolis processeurs *)
                    match goal with 
                      [ |- context c [processorsUpTo (?p') ]] => 
                      let p'':=fresh "p'" in 
                        (set (p'':=p'); 
                          match goal with
                            [|- context c [cons (exist ?f (S ?p') (?p'Proc)) nil ]] => 
                            let Sp:=fresh "S_"p in set (Sp:=(exist f (S p') (p'Proc)))
                          end;
                          repeat 
                            match goal with 
                              [IHp :  forall pProc : p < bsp_p, ?IHpbody |- _] 
                              =>
                              pose(IHp':= IHp (proj2_sig p''));replace (exist (fun pid : nat => pid < bsp_p) p (proj2_sig p''))  with p'' in IHp' by(reflexivity);
                                try (clear IHp)
                                
                            end
                        )
                        end;
  (* séparation de la dernière liste liste  *)
                    repeat  match goal with 
                              [|-context [Flatten.flatten (map (get ?plst) (processorsUpTo ?p ++ [?S_p]))]] =>
                              replace (Flatten.flatten (map (get plst) (processorsUpTo p ++ [S_p]))) 
                                 with ((Flatten.flatten (map (get plst) (processorsUpTo p))) ++ get plst S_p)
                                   by(
                                     rewrite map_app;
                                       rewrite Flatten.flattenApp;
                                         simpl ; rewrite app_nil_r; reflexivity)
                            end)
            ].
  

  Lemma inductionParList :  forall A (P: list A -> Prop) (plst : par (list A)),
    P (get plst firstProcessor)
    -> 
    (forall (p Sp:processor) ,
      (proj1_sig Sp )= S (proj1_sig p) -> 
      P (flatten (map (get plst) (processorsUpTo p))) -> P ( (flatten(map (get plst) (processorsUpTo p))) ++  (get plst Sp)))
    -> P ( listOfParList plst).
  Proof.  
    intros A P plst H H0.
    listOfParListInduction.
    simpl.
    assumption.
    apply H0.
     reflexivity.
   apply IHp'.
Qed.

Lemma inListOfPar: forall (A:Type)(v:par(list A))(a:A),
  In a (listOfParList v) <->exists i:processor, In a (get v i).
Proof.
  unfold listOfParList; intros A v a; split; intro H; [
      rewrite inFlatten in H; destruct H as [l Hl]; rewrite in_map_iff in Hl;
        destruct Hl as [Hl Hl']; destruct Hl as [j Hl]; destruct Hl; exists j; rewrite H; auto |
          rewrite inFlatten; destruct H as [j H]; exists (get v j); split; auto;
            apply in_map_iff; exists j; intuition ].
Qed.


End TYPE.

Module Make  
  (BsmlPrimitives: PRIMITIVES) : 
  TYPE(BsmlPrimitives).
  
  Include TYPE(BsmlPrimitives).

End Make.