Require Import List.
Require Import LIFO.List.InSig.
Require Import LIFO.List.Misc.
Require CertifiedBSML.Primitives.Specification.
Require CertifiedBSML.Primitives.Properties.

Set Implicit Arguments.

Module Type TYPE
  (Bsml : CertifiedBSML.Primitives.Specification.PRIMITIVES)
  (BsmlProp: CertifiedBSML.Primitives.Properties.TYPE(Bsml)).

  Program Definition parSig 
    (A:Type)(v:Bsml.par A)(P:A->Prop)(H:forall i, P(Bsml.get v i)) : 
    { vr : Bsml.par (sig P) | forall i, Bsml.get vr i =  Bsml.get v i} := 
    Bsml.mkpar(fun i=>@exist A P (Bsml.get v i) (H i)).

  Definition lparSig 
    (A:Type)(v:Bsml.par A)(P:A->Prop)(H:forall i, P(Bsml.get v i)) :=
    proj2_sig(parSig v P H).

  Hint Rewrite lparSig : par_sig.

  Program Definition parInSig (A:Type)(v:Bsml.par (list A))(P:A->Prop)
    (H:forall i, forall x, In x (Bsml.get v i)->P x) : 
    { vr : Bsml.par (list (sig P)) | 
      forall i, listProj (Bsml.get vr i) =  Bsml.get v i} := 
    Bsml.mkpar(fun i=>inSig P (Bsml.get v i) (H i)).
  Next Obligation.
    intros; autorewrite with bsml; apply inSigProj.
  Defined.

  Definition lparInSig (A:Type)(v:Bsml.par (list A))(P:A->Prop)
    (H:forall i, forall x, In x (Bsml.get v i)->P x) :=
    proj2_sig(parInSig v P H).

  Hint Rewrite lparInSig : par_sig.

  Program Definition getParInSig: forall (A:Type)(v:Bsml.par (list A))(P:A->Prop)
    (H:forall i, forall x, In x (Bsml.get v i)->P x)(i:Bsml.processor), 
    Bsml.get (parInSig v P H) i = inSig P (Bsml.get v i) (H i).
  Proof.
    intros; simpl; autorewrite with bsml; auto.
  Qed.

  Hint Rewrite getParInSig : par_sig.

End TYPE.
  

Module Make
  (Bsml : CertifiedBSML.Primitives.Specification.PRIMITIVES)
  (BsmlProp : CertifiedBSML.Primitives.Properties.TYPE(Bsml)).

  Include TYPE(Bsml)(BsmlProp).

End Make.