open Bsml
open Stdlib.Base
open Tools
open NatTools
open Specification 
open Datatypes 

module Primitives : PRIMITIVES with type 'a par = 'a Bsml.par  =
struct

  (** Processors are indexed by unary natural  *)
  type processor = nat

  (** Parallel vector are those of BSML  *)
  type 'a par = 'a Bsml.par

  (** bsp_p is the Peano encoding of the number of processor available  *)
  let bsp_p = nat_of_int (Bsml.bsp_p)

  (** get is for specification purpose only, it should not appear in programs   *)
  let get v i = failwith "get is for specification purpose only, it should not be called"

  (** mkpar build a parallel vector  by applying [f] to each processor 
      we could try a version applying replicate of [f] to  a predefined vector of pid (encoded as Peano natural)
  *)
  let mkpar f = Bsml.mkpar (fun i -> f (nat_of_int i))

  (** apply is the point-wise  application of vectors *)
  let apply = Bsml.apply

  (** Communication primitive:
       (processor -> 'a1) par -> (processor -> 'a1) par
      The functions f of the vector vf  are encapuslated using a function translating int into Peano nat
      then the Bsml put is performed and the result is  encapsulated to provide a function giving the incomming data for a given processor(Peano nat).
  *)
  let put vf =
    parfun (fun f -> compose f int_of_nat) 
      (Bsml.put(parfun (fun f -> compose f nat_of_int)vf))

  let proj v = compose (Bsml.proj v) int_of_nat

end
