open Specification
open Datatypes
open Flatten
open InSig
open EqNat
open List0
open MinMax
open Option
open Peano
open NPeano
open Peano_dec
open Heat1D
open NatTools
open BsmlNat.Primitives
open Benchmarking

module Number : NUMBER with type number = float = 
struct
  type number = float 
  let add = ( +. )
  let mul = ( *. )
  let sub = ( -. )
  let div = ( /. )
  let opp = ( ~-. )
end

module Processors = 
struct 
  type __ = Obj.t
  let __ = let rec f _ = Obj.repr f in Obj.repr f
  let max_pid =  nat_of_int  (Bsml.bsp_p - 1)
  let firstProcessor = nat_of_int 0
  let lastProcessor = nat_of_int (Bsml.bsp_p - 1)
  let processorOfNatMod n = modulo n (nat_of_int Bsml.bsp_p)
  let natOfProcessor p =p
  let processorOfNat i = i
  let processorEqDec p1 p2 = eq_nat_dec p1 p2
  let processor_beq p1 p2 = eq_nat_dec p1 p2
  let processors = List.map nat_of_int (Tools.from_to 0 (Bsml.bsp_p-1))
  let processorsUpTo i = List.map nat_of_int (Tools.from_to 0 (int_of_nat i))
  let processorsStartingAt i =  List.map nat_of_int (Tools.from_to (int_of_nat i) (Bsml.bsp_p-1))
  
  let processorsRevInduction_obligation_2 x x0 =
    let rec f n x1 =
      match n with
      | O -> x1
      | S n0 -> let sp = S n0 in f n0 (x n0 sp __ __ x1)
    in f lastProcessor x0

  let processorsRevInduction x x0 =
    processorsRevInduction_obligation_2 x x0

  let listOfParList plst =  failwith "listOfParList is for Specification purpose only" 
end

module ParSig = 
struct
  let parSig v = v 
  let parInSig lv = lv 
end

module Base = Base.Make(BsmlNat.Primitives)(Processors)(ParSig)

module HeatCoq = Heat1D.Make(BsmlNat.Primitives)(Number)(Processors)(ParSig)(Base)
module HeatCoqDefunct = Heat1D_defunct.Make(BsmlNat.Primitives)(Processors)(ParSig)(Base)

(* let heatCoq (leftBound:float) (rightBound:float) (dt:float) (dx:float) (gamma:float) (u:(float list)par) =  *)
(*   HeatCoq.heatEquationPar leftBound rightBound dt dx gamma u *)

open Stdlib.Base

let coq_generator x =  (Random.float 60.) -. 30.
let direct_generator x =  (Random.float 60.) -. 30.



let _ = 
  Benchmarking.program_name:="HeatEquation";
  Random.self_init();
  (* let iter = try int_of_string (Sys.argv.(1)) with _ -> 200 in *)
  (* let size = try int_of_string (Sys.argv.(2)) with _ -> 10000 in *)
  let leftBound = 20. 
  and rightBound = 30. 
  and gamma = 0.1
  (* and f x = (Random.float 60.) -. 30. *)
  in
  benchmarking  
    (fun s u ->let dx =  10. /. (float s) in   let dt = (dx*.dx)/.(10.*.gamma) in HeatCoq.heatEquationPar leftBound rightBound dt dx gamma u) 
    coq_generator
    (fun s u ->let dx =  10. /. (float s) in   let dt = (dx*.dx)/.(10.*.gamma) in  HeatDirect.heat leftBound rightBound dt dx gamma u)
    direct_generator
    (fun s u -> let dx =  10. /. (float s) in let  dt = (dx*.dx)/.(10.*.gamma) in  HeatCoqDefunct.heatEquationPar leftBound rightBound dt dx gamma u) 

