open Bsml
open Stdlib.Base
open Stdlib.Comm

let heatEquationFormula dt dx gamma uI uIMinusOne uIPlusOne =
    gamma *. dt /. (dx *. dx) *. ( uIPlusOne +. uIMinusOne -. uI-. uI ) +. uI

let rec heatEquationSeq leftBound rightBound dt dx gamma (u : float list) : float list =
    match u with 
      | [] -> []
      | uI :: u'  -> 
        match u' with 
          | [] -> [ heatEquationFormula dt dx gamma uI leftBound rightBound ]
          | uIPlusOne:: _ ->
            (heatEquationFormula dt dx gamma uI leftBound uIPlusOne)::
            (heatEquationSeq uI rightBound dt dx gamma u')            

let first l = List.hd l

let last  l = List.hd(List.rev l)

let getBounds l_bound r_bound u =
  let msg = put << fun dst ->
    if dst=($this$+1) && $this$<>(bsp_p-1)
    then [last $u$] 
    else if dst=($this$-1) && $this$<>0
    then [first $u$] 
    else [] >> in
    << if $this$=0
       then l_bound 
       else List.hd ($msg$ ($this$-1)) >>,
    << if $this$=(bsp_p-1) 
       then r_bound 
       else List.hd ($msg$ ($this$+1)) >>
;;

let heat leftBound rightBound dt dx gamma (u : (float list) par ) : (float list) par =
  let leftBounds,rightBounds = getBounds leftBound rightBound u in 
  << heatEquationSeq $leftBounds$ $rightBounds$ dt dx gamma $u$ >>
;;

let rec map_aux f aux l = 
  match l with 
      [] -> aux
    | h::t -> map_aux f ((f h)::aux) t

let map f l = List.rev(map_aux f [] l)

let _ = 
  let iter = try int_of_string (Sys.argv.(1)) with _ -> 200 in
  let size = try int_of_string (Sys.argv.(2)) with _ -> 10000 in
  let leftBound = 20. 
  and rightBound = 30. 
  and gamma = 0.1
  and f x = (Random.float 60.) -. 30.
  and dx = 10. /. (float size) in
  let dt = (dx*.dx)/.(10.*.gamma) in
  let u = mkpar(fun pid -> 
    map f (Tools.from_to (if pid < size mod Bsml.bsp_p then 0 else 1) (size/Bsml.bsp_p))) in
  begin
    Gc.set { (Gc.get ()) with Gc.minor_heap_size = 1024*1024*1024*100 };
    let start_minor_collection_vector= << let stat =Gc.stat() in stat.Gc.minor_collections>> in
    let gc_cost= << ref 0. >> in 
    let cost = << ref 0. >> in
			  Bsml.start_timing();
			  let u' = ref u in 
			  for i=1 to iter do
			    u' := heat leftBound rightBound dt dx gamma !u';
			    Bsml.stop_timing();
			    let cost_vec=Bsml.get_cost() in
			    ignore(<< $cost$:=( $cost_vec$ +. ! ($cost$)) >>);
			    Bsml.start_timing();
			    Gc.full_major();
			    Bsml.stop_timing();
			    ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
			    Bsml.start_timing();
			  done;
			  Bsml.stop_timing();
			  ignore(<< $cost$:=( $Bsml.get_cost()$ +. ! ($cost$)) >>);
			  ignore (<< if ( $this$ = 0 ) then (print_float ! $cost$ ; print_string ";")else () >> );
			  let word_vector= << let stat =Gc.stat() in stat.Gc.minor_words +.  stat.Gc.major_words -. stat.Gc.promoted_words>> in
			  let minor_collection_vector= << let stat =Gc.stat() in stat.Gc.minor_collections>> in
			  ignore (<< if ( $this$ = 0 ) then ( print_float $word_vector$; print_string ";";print_int ($minor_collection_vector$ - $start_minor_collection_vector$); print_string ";") else () >> );
			  Bsml.start_timing();
			  Gc.full_major();
			  Bsml.stop_timing();
			    ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
			    ignore (<< if ( $this$ = 0 ) then ( print_float ! $gc_cost$ )  else () >> )
  end
 
