Require Import List.
Require Import Program.
Require Import LIFO.Algebra.AlgebraClasses.
Set Implicit Arguments.

(** Removing the Some constructor  *)

(** using default value  *)
Definition no_some_default  (A:Type) (a: option A) (d: A) :=
  match a with 
    None => d
    | Some a => a
  end.


(** using proof that it is different from None  *)
Program Definition 
  noSome(A:Type)(a:option A| a <> None) : { x:A | Some x = a } := 
  match a with 
    | None => _ 
    | Some x => x
  end.


Lemma someNoSomeId: 
  forall(A:Type)(a:option A)(H:a<>None),
    Some(proj1_sig(noSome(exist _ a H))) = a.
Proof.
  intros A a Ha;
    destruct a; simpl; auto; elim Ha; auto.
Qed.

Lemma someInjectiveEq: 
  forall(A:Type)(x y:A), 
    x = y <-> Some x = Some y.
Proof.
  intros A x y; split;intro H; [
    f_equal | inversion H ]; auto.
Qed.

Hint Rewrite someNoSomeId: option.


Definition isSome (A:Type)(x:option A) : bool := 
  match x with 
    | None => false 
    | Some _ => true
  end.

Lemma isSomeTrueNotNone:
  forall (A:Type)(x: option A), 
    isSome x = true -> x<>None.
Proof.
  destruct x; intro H; discriminate.
Qed.

Fixpoint mapFilterSome (A B :Type)(f: A->option B)(l:list A) := 
  match l with
    | nil => nil
    | cons h t => 
      match (f h) with 
	| Some x => x::mapFilterSome f t
        | None => mapFilterSome f t
      end
  end.

Lemma mapFilterSomeProp: 
  forall (A B :Type)(f: A->option B)(l:list A),
    map Some (mapFilterSome f l) = filter (@isSome B) (map f l).
Proof.
  induction l.
  auto.
  case_eq (f a);
    try(intros b Hfa);(try intros Hfa); simpl; rewrite Hfa; simpl; f_equal; assumption.
Qed.

Lemma mapFilterSome_ext :
  forall A B l (f f' : A -> option B), 
    (forall a:A, f a = f' a) ->
    mapFilterSome f l =  mapFilterSome f' l.
Proof.
  intros A0 B l f f'  H.
  induction l.
  reflexivity.        
  simpl.
  rewrite H.
  rewrite IHl.
  reflexivity.
Qed.
Implicit Arguments mapFilterSome_ext [A B].

(** ** Combining option  *)

Definition right_option A  (a: option A) (b:option A) := match b with None => a | Some  _ =>  b end.

Global Instance right_option_assoc A : Associative (@right_option A).
Proof.
  constructor. intros x y z.
  destruct x; destruct y; destruct z; reflexivity.
Qed.

Global Instance right_option_neutral A : Neutral (@right_option A) (@None A).
Proof.
  firstorder.
  destruct a; reflexivity.
Qed.


Definition left_option A  (a: option A) (b:option A) := match a with None => b | Some  _ =>  a end.
Global Instance left_option_assoc A : Associative (@left_option A).
Proof.
  constructor. intros x y z.
  destruct x; destruct y; destruct z; reflexivity.
Qed.

Global Instance left_option_neutral A : Neutral (@left_option A) (@None A).
Proof.
  firstorder.
  destruct a; reflexivity.
Qed.

Definition optionify (A B:Type) (op : A -> A ->B) := 
fun a b => 
  match a with
    None => None 
    | Some a' => match b with 
                   None => None
                   | Some b' => Some (op a' b') 
                 end
  end.

Definition optionify_default (A B:Type) (op : A -> A ->B) (d:A) := 
fun a b => 
  match a with
    None => match b with 
                   None => None
                   | Some b' => Some (op d b') 
                 end
    | Some a' => match b with 
                   None => Some (op a' d)
                   | Some b' => Some (op a' b') 
                 end
  end. 

Definition optionify' (A:Type) (op : A -> A -> A) := 
fun a b => 
  match a with
    None => b
    | Some a' => match b with 
                   None => a
                   | Some b' => Some (op a' b') 
                 end
  end.

Definition optionify'_default (A:Type) (op : A -> A -> A) (d:A) := 
fun a b => 
  match a with
    None => match b with 
              None => d
              | Some b' => b'
            end
    | Some a' => match b with 
                   None => a'
                   | Some b' => op a' b'
                 end
  end.
(*  *)