open BsmlSpec
open Datatypes
open EqNat
open List0
open MinMax
open Option
open Peano
open NPeano
open HeatEquation1D
open NatTools
open BsmlNat.Primitives

module Number : NUMBER with type number = float = 
struct
  type number = float 
  let add = ( +. )
  let mul = ( *. )
  let sub = ( -. )
  let div = ( /. )
  let opp = ( ~-. )
end

module Processors = 
struct 
  let firstProcessor = nat_of_int 0
  let lastProcessor = nat_of_int (Bsml.bsp_p - 1)
  let processorOfNatMod n = modulo n (nat_of_int Bsml.bsp_p)
  let processors = List.map nat_of_int (Tools.from_to 0 (Bsml.bsp_p-1))
  let processorsUpTo i = List.map nat_of_int (Tools.from_to 0 (int_of_nat i))
end

module ParSig = 
struct
  let parSig v = v 
  let parInSig lv = lv 
end

module Base = BsmlBase.Make(BsmlNat.Primitives)(Processors)(ParSig)

include HeatEquation1D.Make(BsmlNat.Primitives)(Number)(Processors)(ParSig)(Base)

let heat (leftBound:float) (rightBound:float) (dt:float) (dx:float) (gamma:float) (u:(float list)par) = 
  heatEquationPar leftBound rightBound dt dx gamma u

open Stdlib.Base

let _ = 
  let iter = try int_of_string (Sys.argv.(1)) with _ -> 200 in
  let size = try int_of_string (Sys.argv.(2)) with _ -> 10000 in
  let leftBound = 20. 
  and rightBound = 30. 
  and gamma = 0.1
  and f x = (Random.float 60.) -. 30.
  and dx = 10. /. (float size) in
  let dt = (dx*.dx)/.(10.*.gamma) in
  let u = mkpar(fun pid -> List.map f (Tools.from_to (if pid < nat_of_int (size mod Bsml.bsp_p) then 0 else 1) (size/Bsml.bsp_p))) in
  begin
    Gc.set { (Gc.get ()) with Gc.minor_heap_size = 1024*1024*1024 };
    Bsml.start_timing();
    let u' = ref u in 
    for i=1 to iter do
      u' := heat leftBound rightBound dt dx gamma !u'
    done;
    Gc.full_major();
    Bsml.stop_timing();
    let cost = Bsml.get_cost() in
    ignore (<< if ( $this$ = 0 ) then print_float $cost$ else () >> )
  end
      


