 Unset Automatic Introduction.    (* coq 8.3 *)
Require Import FixedLength.
Require Import BoundedNat.

Require Import List.
Require Import Omega.

Set Implicit Arguments.

(** A strong version of Nth *)

Program Definition strongNth 
  (A:Type) (l:list A) (pos:boundedNat(length l)): 
  {a : A | forall d, nth pos l d = a}.
Proof.
  intros A l pos ; destruct pos as [pos Hpos]; destruct l.
    absurd(pos<0) ; (omega || assumption).
    exists (nth pos (a::l) a); intro d; apply nth_indep; assumption.
Defined.

Program Lemma strongNthIsProofIrrelevant 
  (A:Type)(l :list A)
  (pos1 pos2: boundedNat (length l)):
    proj1_sig pos1 =  proj1_sig pos2 ->
    proj1_sig (strongNth l pos1) = 
    proj1_sig (strongNth l pos2).
Proof.
  intros A l pos1 pos2 H.
  induction l ; simpl in * ; 
    destruct pos1 as [pos1 Hpos1]; destruct pos2 as [pos2 Hpos2].
      absurd(pos1<0); [omega | assumption].
      simpl in *; subst; reflexivity.
Qed.

(* Is it really necessary ? *)

Program Definition boundedStrongNth 
  (A:Type)(len:nat)(l:FixedLength.list A len)
  (pos:boundedNat len): 
  {a : A | forall d, nth pos l d = a} :=
  strongNth l pos.
Next Obligation.
  destruct pos; destruct l; simpl; subst; auto.
Defined.

Program Lemma boundedStrongNthIsProofIrrelevant 
  (A:Type)(len1 len2:nat)
  (l1 : FixedLength.list A len1)
  (l2 : FixedLength.list A len2)
  (pos1 : boundedNat len1)
  (pos2 : boundedNat len2): 
  len1 = len2 ->
  proj1_sig l1 = proj1_sig l2 ->
  proj1_sig pos1 = proj1_sig pos2 ->
    proj1_sig (boundedStrongNth l1 pos1) = 
    proj1_sig (boundedStrongNth l2 pos2).
Proof.
  intros A len1 len2 l1 l2 pos1 pos2 H H0 H1.
  destruct l1 as [l1 Hl1];   destruct l2 as [l2 Hl2]; 
    destruct pos1 as [pos1 Hpos1]; destruct pos2 as [pos2 Hpos2].
  induction len1 ; destruct len2 ; try(discriminate).
    absurd(pos1<0) ; [ omega | assumption ].
    destruct l1.
      discriminate Hl1.
      destruct l2.
        discriminate Hl2.
        simpl in *; inversion H0; subst; reflexivity. 
Qed.

Program Lemma nthStrongNth:
  forall (A:Type) (l:list A)(pos:boundedNat(length l))(default:A),
    nth pos l default = strongNth l pos.
Proof.
  intros A l pos.
  set(a:=strongNth l pos).
  destruct a as [value Property].
  simpl. 
  assumption.
Qed.

Program Lemma strongNthIn : 
  forall (A:Type) (l:list A) (pos:boundedNat(length l))(a:A),
  a = strongNth l pos -> 
  In a l.
Proof.
  intros A l pos x.
  set(a:=strongNth l pos).
  destruct a as [value Property].
  intro H ; subst ; simpl; rewrite <- Property with (d:=value); apply nth_In.
  destruct pos ; auto.
Qed.

Program Lemma inStrongNth : 
  forall (A:Type) (l:list A)(a:A),
  In a l ->
  exists pos, a = strongNth l pos.
Proof.
  intros A l a H.
  induction l. 
    elim H.
    destruct H as [H|H].
    subst.
         assert(0<length(a::l)) by (apply lt_0_Sn);
           exists (exist  _ 0 H: boundedNat (length (a :: l))) ; auto.
      destruct(IHl H) as [pos Heq]; destruct pos as [pos Hpos];
        assert(S pos <length(a0::l)) by (apply lt_n_S;assumption);
          exists(exist (fun n:nat=>n<length(a::l)) (S pos) H0).
      rewrite <- nthStrongNth with (default:=a) in *;
        rewrite Heq; simpl; apply nth_indep; assumption.
Qed.




